#ifdef USE_KIRA

#include "worldline.h"

// #define CAVE 1		// NewN: use potentially-shared-memory malloc
#include "shmem.h"
#include "specks.h"
#include "partiviewc.h"

#include "kira_parti.h"		// declare following as extern "C"


extern struct specklist *get_parti( struct stuff *st, double realtime );

typedef worldbundle *worldbundleptr;
struct worldstuff {
    int nh;
    worldbundleptr *wh;
    int ih;			// current worldbundle index
    real tmin, tmax;
    real tcur;			// current time
    real treq;			// requested time
    struct specklist *sl;
};

int open_parti(struct stuff *st, char *filename, int verbose)
{
    ifstream s(filename);
    if (!s) {
	msg("Data file %s not found.", filename);
	return 0;
    }

    worldbundle *wb;
    struct worldstuff *ww = new worldstuff;
    ww->wh = new worldbundleptr[1024];

    ww->nh = 0;
    while (ww->nh < 1024 && (wb = read_bundle(s, verbose)))
	ww->wh[ww->nh++] = wb;

    if (ww->nh == 0) {
	msg("Data file %s not in worldbundle format.", filename);
	return 0;
    }

    ww->ih = 0;
    ww->tmin = ww->wh[0]->get_t_min();
    ww->tmax = ww->wh[ww->nh-1]->get_t_max();
    ww->sl = NULL;
    ww->tcur = ww->treq = -1.0;		// invalid

    st->dyndata = 1;
    st->dyndatafunc = get_parti;
    st->dyndatadata = ww;

    s.close();
    return 1;
}

double get_parti_time( struct stuff *st ) {
    if(st->dyndatadata == NULL)
	return 0.0;
    return ((struct worldstuff *)(st->dyndatadata))->tcur;
}

void set_parti_time( struct stuff *st, double reqtime ) {
    if(st->dyndatadata == NULL)
	return;
    // otherwise please seek to requested time ...
}

int get_parti_time_range( struct stuff *st, double *tmin, double *tmax )
{
    struct worldstuff *ww = (struct worldstuff *)st->dyndatadata;
    if(ww == NULL) {
	*tmin = 0, *tmax = 1;
	return 0;
    }

    *tmin = ww->tmin;
    *tmax = ww->tmax;
    return 1;
}


    // speck fields:
    //  [0] worldline index (= kira index for singles, - and unique for others)
    //  [1] mass
    //  [2] number in clump
    //  [3] top-level name
    //  [4] address in clump
#define NDATAFIELDS  5

speck *assign_specks(dyn *b, dyn *top, int addr, speck *sp, specklist *sl)
{
    // Convert relative coordinates to absolute.

    if (b != top)
	b->inc_pos(b->get_parent()->get_pos());

    sp->p.x[0] = b->get_pos()[0];
    sp->p.x[1] = b->get_pos()[1];
    sp->p.x[2] = b->get_pos()[2];

    sp->val[0] = b->get_index();
    if (!b->is_leaf())
	sp->val[0] = -sp->val[0];	// distinguish CM

    sp->val[1] = b->get_mass();

    sp->val[2] = 0;			// complete (add n_leaves) in to_parti

    // Offset unperturbed binaries by 100 in field 2.

    if (b != top) {
    	dyn *bb = b;
    	while (bb->get_elder_sister())
    	    bb = bb->get_elder_sister();
    	if (bb->get_kepler()) sp->val[2] += 100;
    }

    sp->val[3] = top->get_index();

    // Addr is 0 for top-level leaf, 1 for top-level CM,
    // constructed from parent addr otherwise.

    sp->val[4] = addr;

    sl->nspecks++;
    sp = NextSpeck(sp, sl, 1);

    // Recursion.

    addr *= 2;
    for_all_daughters(dyn, b, bb)
	sp = assign_specks(bb, top, addr++, sp, sl);

    return sp;
}


struct specklist *to_parti(dyn *root, struct stuff *st, struct specklist *sl)
{
    if(sl == NULL) {
	// Use NewN to allocate these in (potentially) shared memory,
	// rather than new which will allocate in local malloc pool.
	// Needed for virdir/cave version.
        sl = NewN( struct specklist, 1 );
	memset(sl, 0, sizeof(*sl));
	sl->bytesperspeck = SMALLSPECKSIZE(NDATAFIELDS);
	sl->specks = NewNSpeck(sl, 2*root->n_leaves());
	sl->scaledby = 1;
    }
    struct speck *sp = sl->specks;

    sl->nspecks = 0;
    for_all_daughters(dyn, root, b) {
	int ns = sl->nspecks;
	speck *tsp = sp;
	sp = assign_specks(b, b, !b->is_leaf(), sp, sl);

	int nleaves = (sl->nspecks - ns + 1) / 2;	// = b->n_leaves() !!
	ns = sl->nspecks - ns;				// = #specks in tree
	for( ; ns > 0; ns--) {
	    tsp->val[2] += nleaves;
	    tsp = NextSpeck(tsp, sl, 1);
	}
    }

    return sl;
}

struct specklist *get_parti( struct stuff *st, double realtime )
{
    struct worldstuff *ww = (struct worldstuff *)(st->dyndatadata);

    if (!ww)
	return NULL;


    ww->treq = realtime;

    if (realtime < ww->tmin) realtime = ww->tmin;
    if (realtime > ww->tmax) realtime = ww->tmax;

    if (ww->tcur == realtime)
	return ww->sl;

    int ih = ww->ih;
    int nh = ww->nh;

    worldbundle *wb = ww->wh[ih];
    for(; realtime > wb->get_t_max() && ih < nh-1; ih++, wb = ww->wh[ih])
	;
    for(; realtime < wb->get_t_min() && ih > 0; ih--, wb = ww->wh[ih])
	;

    dyn *root = create_interpolated_tree(wb, realtime);
    ww->sl = to_parti(root, st, ww->sl);
    rmtree(root);

    ww->ih = ih;
    ww->tcur = realtime;

    return ww->sl;
}

#endif /*USE_KIRA*/
